% plot a tradeoff curves for the digital circuit sizing example
% where Amax and Pmax are allowed to vary.
%
% this is an example taken directly from the paper:
%
%   A Tutorial on Geometric Programming (see pages 25-29)
%   by Boyd, Kim, Vandenberghe, and Hassibi.
%
% another treatment of this example and its dynamic programming
% formulation (which is used for timing constraints) is in the paper:
%
%   Digital Circuit Optimization via Geometric Programming
%   by Boyd, Kim, Patil, and Horowitz.
%   (see pages 5-14)

% digital circuit shown in figure 4 (page 28) of GP tutorial paper
m = 7;  % number of cells
n = 8;  % number of edges
A = sparse(m,n);

% A is standard cell-edge incidence matrix of the circuit
% A_ij = 1 if edge j comes out of cell i, -1 if it comes in, 0 otherwise
  A(1,1) =     1;
  A(2,2) =     1;
  A(2,3) =     1;
  A(3,4) =     1;
  A(3,8) =     1;
  A(4,1) =    -1;
  A(4,2) =    -1;
  A(4,5) =     1;
  A(4,6) =     1;
  A(5,3) =    -1;
  A(5,4) =    -1;
  A(5,7) =     1;
  A(6,5) =    -1;
  A(7,6) =    -1;
  A(7,7) =    -1;
  A(7,8) =    -1;

% decompose A into positive and negative part 
Ap = double(A > 0);
An = double(A < 0);

% constants
f = [1 0.8 1 0.7 0.7 0.5 0.5]';
e = [1 2 1 1.5 1.5 1 2]';
Cout6 = 10;
Cout7 = 10;

a     = ones(m,1);
alpha = ones(m,1);
beta  = ones(m,1);
gamma = ones(m,1);

% optimization variables
gpvar x(m)                 % sizes
gpvar t(m)                 % arrival times

% optimization relations and constraints
cin = alpha + beta.*x;     % input capacitance
cload = Ap*An'*cin;        % load capacitance
cload(6) = Cout6;          % Cload of output gate 6
cload(7) = Cout7;          % Cload of output gate 7
d = cload./x;              % delay

power = (f.*e)'*x;         % power
area = a'*x;               % area

% problem constraints
constr_x = ones(m,1) <= x; % all sizes greater than 1 (normalized)

% create timing constraints
constr_timing = Ap'*t + An'*d <= An'*t;
input_timing  = d(1:3) <= t(1:3);

% objective is the upper bound on the overall delay
D = max(t(6),t(7));
constr = [power <= 1; area <= 1; ... % just a place holders
                                     % these constraints will change
          constr_timing; input_timing; constr_x];

% varying parameters for an optimal trade-off curve
N = 10;
Pmax = linspace(10,100,N);
Amax = [25 50 100];
min_delay = zeros(length(Amax),N);

for k = 1:length(Amax)
  for n = 1:N
    % add constraints that have varying parameters
    constr(1) = power <= Pmax(n);
    constr(2) = area  <= Amax(k);

    % solve the GP problem and compute the optimal volume
    [obj_value, solution, status] = gpsolve(D, constr);
    min_delay(k,n) = obj_value;
  end
end

plot(Pmax,min_delay(1,:), Pmax,min_delay(2,:), Pmax,min_delay(3,:));
xlabel('Pmax'); ylabel('Dmin');
% print -depsc dig_ckt_sizing_tradeoff
